
var POOL_NAMES = 
	['i', 'ii', 'iii', 'iv', 'v', 'vi', 'vii', 'viii', 'ix', 'x', 'xi', 'xii' ];

BC_1_8 =   ['AD001', 'AD002', 'AD003', 'AD004', 'AD005', 'AD006', 'AD007', 'AD008'];
BC_8_1 =   ['AD008', 'AD007', 'AD006', 'AD005', 'AD004', 'AD003', 'AD002', 'AD001'];
BC_9_16 =  ['AD009', 'AD010', 'AD011', 'AD012', 'AD013', 'AD014', 'AD015', 'AD016'];
BC_16_9 =  ['AD016', 'AD015', 'AD014', 'AD013', 'AD012', 'AD011', 'AD010', 'AD009'];
BC_18_27 = ['AD018', 'AD019', 'AD020', 'AD021', 'AD022', 'AD023', 'AD025', 'AD027'];
BC_27_18 = ['AD027', 'AD025', 'AD023', 'AD022', 'AD021', 'AD020', 'AD019', 'AD018'];

// Index Sets = IS
var IS_G1 = {color: 'bg-green',  barcodes: BC_1_8};
var IS_G2 = {color: 'bg-green',  barcodes: BC_9_16};
var IS_G3 = {color: 'bg-green',  barcodes: BC_18_27};

var IS_G4 = {color: 'bg-green-dk',  barcodes: BC_8_1};
var IS_G5 = {color: 'bg-green-dk',  barcodes: BC_16_9};
var IS_G6 = {color: 'bg-green-dk',  barcodes: BC_27_18};

var IS_B1 = {color: 'bg-blue',   barcodes: BC_1_8};
var IS_B2 = {color: 'bg-blue',   barcodes: BC_9_16};
var IS_B3 = {color: 'bg-blue',   barcodes: BC_18_27};

var IS_B4 = {color: 'bg-blue-dk',   barcodes: BC_8_1};
var IS_B5 = {color: 'bg-blue-dk',   barcodes: BC_16_9};
var IS_B6 = {color: 'bg-blue-dk',   barcodes: BC_27_18};

var IS_Y1 = {color: 'bg-yellow', barcodes: BC_1_8};
var IS_Y2 = {color: 'bg-yellow', barcodes: BC_9_16};
var IS_Y3 = {color: 'bg-yellow', barcodes: BC_18_27};

var IS_Y4 = {color: 'bg-yellow-dk', barcodes: BC_8_1};
var IS_Y5 = {color: 'bg-yellow-dk', barcodes: BC_16_9};
var IS_Y6 = {color: 'bg-yellow-dk', barcodes: BC_27_18};


var POOL_SCHEMA =
[
	{
		id: '4by24',
		name: '4 pools × 24 samples',
 		numPools: 4,
 		numWellsPerPool: 24,
 		columnClasses: ['pool-left', 'pool-middle', 'pool-right'],
		barcodeVariants:
			[
			 	{
			 		name: 'A-B-C-D', 
			 		indexSets: [IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y6, IS_G4, IS_B5, IS_Y3, IS_G4, IS_B5, IS_Y6]
			 	},
			 	{
			 		name: 'A-A-A-A', 
			 		indexSets: [IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3]
			 	},
			 	{
			 		/* Same as A-A-A-A, kept for backwards compatibility */
			 		name: 'FULL', 
			 		disabled: 1,
			 		indexSets: [IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3, IS_G1, IS_B2, IS_Y3]
			 	}
			],

		getPoolNumForColumn: function(colNum)
		{
			var poolNum = Math.floor(colNum / 3);
			return poolNum;
		},

 		getClassNameForWell: function(well)
		{
			var c = well.column;
			var cls = 'primary ';
			cls += this.columnClasses[c % 3];
			return cls;
		},
		
		getWellsInPool: function(plate, poolNum)
		{
			var result = [];
			var index = poolNum * this.numWellsPerPool;
			while (result.length < this.numWellsPerPool && index < plate.wells.length)
			{
				result[result.length] = plate.wells[index];
				index++;
			}
			return result;
		}
	
	},
	{
		id: '4by16',
		name: '4 pools × 16 samples',
 		numPools: 4,
 		numWellsPerPool: 16,
 		columnClasses: ['pool-left', 'pool-right'],
		barcodeVariants:
			[
			 	{name: 'A-B-C-A', indexSets: [null, null, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, null, null]},
			 	{name: 'B-C-A-B', indexSets: [null, null, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, null, null]},
			 	{name: 'C-A-B-C', indexSets: [null, null, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, null, null]}
			],

		getPoolNumForColumn: function(colNum)
		{
			var poolNum = -1;
			if (colNum >= 2 && colNum <= 9)
			{
				poolNum = Math.floor((colNum - 2) / 2);
			}
			return poolNum;
		},
 		
 		getClassNameForWell: function(well)
		{
			var c = well.column;
			var cls = (c >= 2 && c <= 9) ? 'primary ' : 'secondary ';
			cls += this.columnClasses[c % 2];
			return cls;
		},
		
		getWellsInPool: function(plate, poolNum)
 		{
 			var result = [];
 			var index = poolNum*this.numWellsPerPool + 2 * plate.rows;
			while (result.length < this.numWellsPerPool && index < plate.wells.length)
			{
				result[result.length] = plate.wells[index];
				index++;
			}
 			return result;
 		}
 	},
	{
		id: '5by16',
		name: '5 pools × 16 samples',
		numPools: 5,
		numWellsPerPool: 16,
		columnClasses: ['pool-right', 'pool-left'],
		barcodeVariants:
			[
			 	{name: 'A-B-C-A-B', indexSets: [null, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, null]},
			 	{name: 'B-C-A-B-C', indexSets: [null, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, null]},
			 	{name: 'C-A-B-C-A', indexSets: [null, IS_Y2, IS_Y3, IS_G1, IS_G2, IS_B3, IS_B1, IS_Y2, IS_Y3, IS_G1, IS_G2, null]}
			],
	
		getPoolNumForColumn: function(colNum)
		{
			var poolNum = -1;
			if (colNum >= 1 && colNum <= 10)
			{
				poolNum = Math.floor((colNum-1) / 2);
			}
			return poolNum;
		},
 		
		getClassNameForWell: function(well)
		{
			var c = well.column;
			var cls = (c >= 1 && c <= 10) ? 'primary ' : 'secondary ';
			cls += this.columnClasses[c % 2];
			return cls;
		},
		
		getWellsInPool: function(plate, poolNum)
 		{
 			var result = [];
 			var index = poolNum*this.numWellsPerPool+plate.rows;
			while (result.length < this.numWellsPerPool && index < plate.wells.length)
			{
				result[result.length] = plate.wells[index];
				index++;
			}
 			return result;
 		}
	}
];

/**
	Methods for handling pooling layout schemas.
*/
var PoolSchema = function()
{
	var ps = {};
	
	ps.initList = function(list, selectedId)
	{
		if (!selectedId) selectedId = POOL_SCHEMA[0].id;
		var selectedIndex = 0;
		for (var i = 0; i < POOL_SCHEMA.length; i++)
		{
			var schema = POOL_SCHEMA[i];
			var option = new Option(schema.name, schema.id);
			list[list.length] = option;
			if (selectedId == schema.id) selectedIndex = i;
		}
		list.selectedIndex = selectedIndex;
		return POOL_SCHEMA[selectedIndex];
	}
	
	ps.initVariantList = function(list, schema, selectedId)
	{
		if (!selectedId) selectedId = schema.barcodeVariants[0].name;
		var selectedIndex = 0;
		list.length = 0;
		for (var i = 0; i < schema.barcodeVariants.length; i++)
		{
			var variant = schema.barcodeVariants[i];
			if (!variant.disabled)
			{
				var option = new Option(variant.name, variant.name);
				list[list.length] = option;
				if (selectedId == variant.name) selectedIndex = i;
			}
		}
		list.selectedIndex = selectedIndex;
		return schema.barcodeVariants[selectedIndex];
	}
	
	ps.getById = function(schemaId)
	{
		for (var i = 0; i < POOL_SCHEMA.length; i++)
		{
			if (POOL_SCHEMA[i].id == schemaId)
			{
				return POOL_SCHEMA[i];
			}
		}
		return null;
	}
	
	ps.getBarcodeVariantByName = function(schema, barcodeVariant)
	{
		for (var i = 0; i < schema.barcodeVariants.length; i++)
		{
			if (barcodeVariant == schema.barcodeVariants[i].name) 
			{
				return schema.barcodeVariants[i];
			}
		}
		return schema.barcodeVariants[0];
	}
	
	ps.buildPoolTableRow = function(schema, columns, interactive)
	{
		var html = '<th></th>';
		var colNum = 0;
		var poolClass = interactive ? 'pool link' : 'pool';
		var afterPool = false;
		while (colNum < columns)
		{
			var poolNum = schema ? schema.getPoolNumForColumn(colNum) : -1;
			var colspan = 1;
			if (poolNum == -1)
			{
				while (colNum+colspan < columns)
				{
					if (schema.getPoolNumForColumn(colNum+colspan) != -1) break;
					colspan++;
				}
				var className = afterPool ? 'nopool-after-pool' : 'nopool-before-pool';
				html += '<th class="'+className+'" colspan="'+colspan+'">&nbsp;</th>';
			}
			else
			{
				while (colNum+colspan < columns)
				{
					if (schema.getPoolNumForColumn(colNum+colspan) != poolNum) break;
					colspan++;
				}
				html += '<th id="pool.' + poolNum + '" colspan="'+colspan+'"';
				html += ' class="'+poolClass+'"';
				html += ' data-pool-num="' + poolNum + '"';
				html += '>'+POOL_NAMES[poolNum]+'</th>';
				afterPool = true;
			}
			colNum += colspan;
		}
		
		var row = document.getElementById('pool-row');
		row.innerHTML = html;
	}
	
	return ps;
}();


var PoolMix = function()
{
	
	var LIMIT_FOR_EXTRA_LARGE_MIX = 1.0;
	var AVAILABLE_VOLUME = 10.0;
	var AVAILABLE_VOLUME_AFTER_SPEEDVAC = 5.0;
	
	var pm = {};
	
	/**
		Round the value to the nearest value with 
		the given number of decimals (default=1).
	*/
	pm.round = function(value, numDecimals)
	{
		if (!numDecimals) numDecimals = 1;
		var pow = Math.pow(10, numDecimals);
		return Math.round(value * pow) / pow;
	}
	
	/**
		Round the value down to the nearest value with 
		the given number of decimals (default=1).
	*/
	pm.floor = function(value, numDecimals)
	{
		if (!numDecimals) numDecimals = 1;
		var pow = Math.pow(10, numDecimals);
		return Math.floor(value * pow) / pow;
	}
	
	/**
		Round the value up to the nearest value with 
		the given number of decimals (default=1).
	*/
	pm.ceil = function(value, numDecimals)
	{
		if (!numDecimals) numDecimals = 1;
		var pow = Math.pow(10, numDecimals);
		return Math.ceil(value * pow) / pow;
	}
	
	/**
		Checks if v1 is less than v2 within the given accuracy
		Eg. v1+accuracy <= v2
	 */
	pm.lessThan = function(v1, v2, accuracy)
	{
		return v1 + accuracy < v2;
	}
	
	/**
		Calculate the volume of a lib to use for a pool given the
		target molarity for the pool and average volume for each lib
		in the pool. This method calculates two different volumes:
		
		lib.volume = The theoretical volume needed to reach the same
		amount of DNA as the target molarity and volume suggests
		
		lib.actualVolume: The actual volume that can be used due to
		remaining quantity, lib molarity and mixing strategy
		
		The volume is rounded to one decimal in µl.
		
	*/
	pm.calculateLibVolume = function(lib, targetMolarity, targetVolume, mixingStrategy)
	{
		var volume = pm.round(targetVolume * targetMolarity / lib.molarity);
		var mixFactor = 1;
		
		if (pm.lessThan(volume, LIMIT_FOR_EXTRA_LARGE_MIX, 0.05))
		{
			// Mix 'dil' so that we take about 2 times the minimal limit
			mixFactor = Math.max(2 * LIMIT_FOR_EXTRA_LARGE_MIX / volume);
			// Re-calculate the volume to get the rounding correct (eg. to avoid doubling rounding errors)
			volume = pm.round(mixFactor * targetVolume * targetMolarity / lib.molarity) / mixFactor;
		}
		
		// This is the theoretical volume that should be used...
		lib.volume = volume;
		lib.eb = null;
		lib.mixFactor = mixFactor;
			
		// However, reality has limitations due to remaining volume, mixing strategy, etc.
		// Check what volume that can actually be used
		if (mixingStrategy == 'fixed')
		{
			// Do not use more than target volume or remaining volume
			volume = Math.min(volume, targetVolume, lib.remainingVolume);
		}
		else
		{
			volume = Math.min(volume, lib.remainingVolume);
		}
		lib.actualVolume = volume;			
	}
	
	/**
		Same as above method except that we already know the mixFactor and actualVolume of
		the lib since that has already been saved to the database.
	*/
	pm.calculateLibVolumeForProtocol = function(lib, targetMolarity, targetVolume, mixingStrategy)
	{
		if (lib.mixFactor > 1)
		{
			lib.volume = pm.round(lib.mixFactor * targetVolume * targetMolarity / lib.molarity) / lib.mixFactor;
		}
		else
		{
			lib.volume = pm.round(targetVolume * targetMolarity / lib.molarity);
		}
	}

	
	/**
		Calculate the volume of EB to mix with a lib.
		
		The volume is rounded to one decimal in µl.
	*/
	pm.calculateEbVolume = function(lib, targetMolarity, targetVolume, mixingStrategy)
	{
		var eb = 0;
		if (mixingStrategy == 'fixed')
		{
			eb = targetVolume - lib.volume;
		}
		else if (lib.mixFactor > 1)
		{
			eb = (lib.targetVolume ? lib.targetVolume : targetVolume) - lib.volume;
		}
		
		// This is the theoretical volume that should be used...
		lib.eb = eb;
		lib.mixingStrategy = mixingStrategy;
		
		// However, reality has limitations due to remaining volume, mixing strategy, etc.
		// Check what volume that can actually be used
		if (mixingStrategy == 'fixed' || lib.mixFactor > 1)
		{
			if (pm.lessThan(lib.molarity, targetMolarity, 0.01))
			{
				// Do not dilute if molarity is lower than target molarity
				eb = 0;
			}
			else if (pm.lessThan(lib.actualVolume, lib.volume, 0.05))
			{
				// Reduce EB by the same factor as we have to reduce the volume
				// so the final molarity is still 
				eb = pm.round(eb * lib.actualVolume / lib.volume);
			}
		}		
		lib.actualEb = eb;
	}

	
	
	/**
		Summarize information for the pool. This will calculate the following and
		return as an object:
		
		libVolume: Total volume of libs in the pool (µl)
		libAmount: Total amount of libs in the pool (nano-mol)
		ebVolumeFromLibs: Total EB volume from mixed directly with libs
		ebVolumeExtra: Extra EB to add to the pool (dynamic mixing)
		molarity: The final pool molarity after mixing everything
	*/
	pm.calculateFinalPoolInfo = function(libs, targetMolarity, targetVolume, mixingStrategy)
	{
		var libVolume = 0;
		var libAmount = 0;
		var ebVolumeFromLibs = 0;
		
		for (var libNo = 0; libNo < libs.length; libNo++)
		{
			var lib = libs[libNo];
			
			var volume = lib.actualVolume;
			var eb = lib.actualEb;

			libVolume += volume;
			ebVolumeFromLibs += eb;
			libAmount += volume * lib.molarity;
		}
		
		var info = {};
		info.libVolume = libVolume;
		info.libAmount = libAmount;
		info.ebVolumeFromLibs = ebVolumeFromLibs;
		
		// Calculate extra EB and total pool volume
		if (mixingStrategy == 'dynamic')
		{
			// Use the target molarity to calculate the total volume
			var targetVolume = pm.round(libAmount / targetMolarity);
			info.ebVolumeExtra = pm.round(targetVolume - info.libVolume - info.ebVolumeFromLibs);
			if (info.ebVolumeExtra < 0)
			{
				info.totalVolume = info.libVolume + info.ebVolumeFromLibs;
			}
			else
			{
				info.totalVolume = targetVolume;
			}
		}
		else
		{
			// No extra EB
			info.ebVolumeExtra = 0;
			info.totalVolume = info.libVolume + info.ebVolumeFromLibs;
		}
		// Calcualate the final molarity (which may be different that target due to rounding)
		info.molarity = info.libAmount / info.totalVolume;

		return info;
	}
	
	return pm;
}();
